from __future__ import annotations

import pandas as pd

from .STDataset import STDataset


def low_variance_filter(dataset: STDataset, var: float = 0) -> STDataset:
    """
    Filter genes with low variance. This may also filter genes with 0 expressions.

    Parameters
    ==========
    dataset : STDataset
        STDataset to be filtered.

    var : int, default 0
        Only remain genes with variance greater than (but not equal to) var.

    Returns
    =======
    dataset : STDataset
        A new STDataset instance with filtered genes.
    """
    var_series = pd.Series()
    for gene in dataset.count_df.columns:
        try:
            var_series[gene] = dataset.count_df[gene].sparse.to_dense().var()
        except AttributeError:
            var_series[gene] = dataset.count_df[gene].var()
    var_series = var_series[var_series > var]
    return_dset = STDataset(
        dataset.count_df.reindex(columns=var_series.index),
        dataset.coordinate_df,
        check_duplicate_genes=False,
        sort_spots=False,
    )
    return_dset._array_coordinate = dataset._array_coordinate
    return return_dset


def high_expression_filter(
    dataset: STDataset,
    max_ratio: float = 0.99,
) -> STDataset:
    """
    Filter genes with high expression ratio.

    Parameters
    ==========
    dataset : STDataset
        STDataset to be filtered.

    max_ratio : int, default 0.99
        Only remain genes with expression ratio less than (but not equal to)
        max_ratio.

    Returns
    =======
    dataset : STDataset
        A new STDataset instance with filtered genes.
    """
    temp_df = dataset.count_df.where(dataset.count_df < 1, 1)
    temp_series = temp_df.sum() / temp_df.shape[0]
    drop_genes = []
    for i in temp_series.index:
        if temp_series[i] > max_ratio:
            drop_genes.append(i)
    count_df = dataset.count_df.drop(columns=drop_genes)
    return_dset = STDataset(
        count_df,
        dataset.coordinate_df,
        check_duplicate_genes=False,
        sort_spots=False,
    )
    return_dset._array_coordinate = dataset._array_coordinate
    return return_dset


def quantile_filter(dataset: STDataset, quantile: float = 0.99) -> STDataset:
    """
    Filter genes with quantile.

    Parameters
    ==========
    dataset : STDataset
        STDataset to be filtered.

    quantile : int, default 0.99
        Only remain genes with mean less than (but not equal to) quantile number.

    Returns
    =======
    dataset : STDataset
        A new STDataset instance with filtered genes.
    """
    mean_series = dataset.count_df.mean()
    q = mean_series.quantile(quantile)
    remain_genes = mean_series[mean_series < q].index
    count_df = dataset.count_df.reindex(columns=remain_genes)
    return_dset = STDataset(
        count_df,
        dataset.coordinate_df,
        check_duplicate_genes=False,
        sort_spots=False,
    )
    return_dset._array_coordinate = dataset._array_coordinate
    return return_dset



from __future__ import annotations
from argparse import ArgumentParser
from multiprocessing import cpu_count
from pathlib import Path

from svgbit import load_10X, load_anndata_h5, load_table, run, plot
from svgbit.filters import low_variance_filter, quantile_filter
from svgbit.normalizers import logcpm_normalizer, cpm_normalizer


def main() -> None:
    parser = ArgumentParser(
        prog="svgbit",
        description='''Find spatial variable genes for Spatial Trasncriptomics
                       data.''',
    )
    parser.add_argument(
        "read_path",
        help='''Read Spatial Transcriptomics data. Support format in 10X Space
                Ranger(dir named ``outs``) and anndata hdf5''',
    )
    parser.add_argument(
        "--normalization",
        default=None,
        help='''apply which normalization on read data. If None (default), neither
                normalization will apply. Supported values: None, cpm, logcpm
                (default: %(default)s)''',
    )
    parser.add_argument(
        "--k",
        type=int,
        default=6,
        help='''number of nearest neighbors for KNN network
                (default: %(default)s)''',
    )
    parser.add_argument(
        "--n_svgs",
        type=int,
        default=1000,
        help='''number of SVGs to find clusters (default: %(default)s)''',
    )
    parser.add_argument(
        "--n_svg_clusters",
        type=int,
        default=8,
        help='''number of SVG clusters to find (default: %(default)s)''',
    )
    parser.add_argument(
        "--he_image",
        default=None,
        help='''path to H&E image. Only used for visualization
                (default: %(default)s)''',
    )
    parser.add_argument(
        "--save_dir",
        default=".",
        help='''path to save results (default: %(default)s)''',
    )
    parser.add_argument(
        "--cores",
        type=int,
        default=cpu_count(),
        help='''number of threads to run svgbit (default: %(default)s)''',
    )
    args = parser.parse_args()

    read_path = Path(args.read_path)
    if read_path.is_dir():
        load_func = load_10X
        kwargs = {}
    else:
        if read_path.suffix in [".h5", ".hdf5", ".h5ad"]:
            load_func = load_anndata_h5
            kwargs = {}
        else:
            load_func = load_table
            if ".csv" in read_path.suffixes:
                sep = ","
            elif ".tsv" in read_path.suffixes:
                sep = "\t"
            kwargs = {"index_col": 0, "header": 0, "sep": sep}

    d = load_func(read_path, **kwargs)
    d = low_variance_filter(d)
    d = quantile_filter(d, 0.99)
    norm = args.normalization
    if norm is not None:
        norm = norm.lower()
        if norm == "none":
            pass
        elif norm == "logcpm":
            d = logcpm_normalizer(d)
        elif norm == "cpm":
            d = cpm_normalizer(d)
        else:
            raise ValueError(f"Not supported normalization: {args.normalization}")
    d = run(
        d,
        k=args.k,
        n_svgs=args.n_svgs,
        n_svg_clusters=args.n_svg_clusters,
        cores=args.cores,
    )

    savedir = Path(args.save_dir)
    if not savedir.exists():
        savedir.mkdir()
    d.hotspot_df.to_csv(Path.joinpath(savedir, "hotspot_df.csv"))
    d.AI.to_csv(Path.joinpath(savedir, "AI.csv"))
    d.Di.to_csv(Path.joinpath(savedir, "Di.csv"))
    d.svg_cluster.to_csv(Path.joinpath(savedir, "svg_cluster.csv"))

    plot.svg_heatmap(d, Path.joinpath(savedir, "heatmap.jpg"), args.he_image)
    plot.spot_type_map(d, Path.joinpath(savedir, "type_map.jpg"),
                       args.he_image)


if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from collections import Counter
from copy import deepcopy
from typing import Optional, Tuple, Union
from pathlib import Path

import numpy as np
import pandas as pd
from libpysal.weights import KNN
from libpysal.weights import W as libpysal_W

from . import cluster, density, moran

DataFrames = Union[pd.DataFrame, np.ndarray, Path, str]


class STDataset(object):
    """
    STDataset: A meta class for discribing Spatial Transcriptomics data.

    Parameters
    ==========
    count_df : np.ndarray, pd.DataFrame, str or Path
       Expression matrix for Spatial Transcriptomics Data. If ``str`` or ``Path``
       is given, svgbit will try to read file with given path with pandas.

       Default shape: (spot * gene)

    coordinate_df : np.ndarray, pd.DataFrame, str or Path
       Coordinates for Spatial Transcriptomics Data. If ``str`` or ``Path``
       is given, svgbit will try to read file with given path with pandas.

       Default shape: (spot * 2)

    count_transpose : bool, default False
        Whether to transpose count matrix.

    coordinate_transpose : bool, default False
        Whether to transpose coordinate dataframe.

    count_df_kwargs : dict, default {}
        Keyword arguments pass to ``pandas.read_csv`` if ``str`` or ``Path`` is
        given to ``count_df``.

    coordinate_df_kwargs : dict, default {}
        Keyword arguments pass to ``pandas.read_csv`` if ``str`` or ``Path`` is
        given to ``coordinate_df``.

    make_sparse : bool, default True
        Whether to use sparse DataFrame in order to save memory.

    check_duplicate_genes : bool, default True
        Whether to check duplicated gene names.

    sort_spots : bool, default True
        Whether to sort spots with spots' name.
    """
    def __init__(
        self,
        count_df: DataFrames,
        coordinate_df: DataFrames,
        count_transpose: bool = False,
        coordinate_transpose: bool = False,
        count_df_kwargs: dict = {},
        coordinate_df_kwargs: dict = {},
        make_sparse: bool = True,
        check_duplicate_genes: bool = True,
        sort_spots: bool = True,
    ) -> None:

        # attributes initial
        self._count_df: Optional[pd.DataFrame] = None
        self._coordinate_df: Optional[pd.DataFrame] = None
        self._normalizer: Optional[str] = None
        self._weight: Optional[libpysal_W] = None
        self._weight_type: Tuple[Optional[str], Optional[str]] = (None, None)
        self._hotspot_df: Optional[pd.DataFrame] = None
        self._local_moran_i: Optional[pd.DataFrame] = None
        self._local_moran_p: Optional[pd.DataFrame] = None
        self._AI: Optional[pd.Series] = None
        self._Di: Optional[pd.DataFrame] = None
        self._svg_cluster: Optional[pd.Series] = None
        self._spot_type: Optional[pd.DataFrame] = None
        self._array_coordinate: Optional[pd.DataFrame] = None

        # dataframes check
        if isinstance(count_df, pd.DataFrame):
            self._count_df = deepcopy(count_df)
        elif isinstance(count_df, np.ndarray):
            self._count_df = pd.DataFrame(count_df)
        else:
            self._count_df = pd.read_csv(count_df, **count_df_kwargs)
        if count_transpose:
            self._count_df = self._count_df.T

        if isinstance(coordinate_df, pd.DataFrame):
            self._coordinate_df = deepcopy(coordinate_df)
        elif isinstance(coordinate_df, np.ndarray):
            self._coordinate_df = pd.DataFrame(coordinate_df)
        else:
            self._coordinate_df = pd.read_csv(
                coordinate_df,
                **coordinate_df_kwargs,
            )
        if coordinate_transpose:
            self._coordinate_df = self._coordinate_df.T

        if sort_spots:
            self._count_df.sort_index(inplace=True)
            self._coordinate_df = self._coordinate_df.reindex(
                index=self.count_df.index)
        self._coordinate_df.columns = ["X", "Y"]

        err = "Expression matrix and coordinate file have different number of spots."
        assert self._count_df.shape[0] == self._coordinate_df.shape[0], err
        err = "Spots' name mismatch!"
        assert all(self._count_df.index == self._coordinate_df.index), err

        self._count_df.fillna(0, inplace=True)

        # Rename duplicated columns
        if check_duplicate_genes:
            genes = []
            flag = 0
            c = Counter(self._count_df.columns)
            gene_suffix = {}
            for gene_name in self._count_df.columns:
                if c[gene_name] > 1:
                    flag = 1
                    if gene_name in gene_suffix:
                        gene_suffix[gene_name] += 1
                        gene_name += f".{gene_suffix[gene_name]}"
                    else:
                        gene_suffix[gene_name] = 0
                genes.append(gene_name)
            if flag:
                self._count_df.columns = genes
                print("Duplicated column names found. Auto rename.")
                warnings.warn("Duplicated column names found. Auto rename.")

        if make_sparse:
            self.to_sparse()

    def __repr__(self) -> str:
        descr = f"STDataset with n_spots x n_genes = {self.n_spots} x {self.n_genes}"
        descr = f"{descr}\nApplied normalizers: {self._normalizer}"
        descr = f"{descr}\nAssigned attributes: "
        flag = 0
        for attr in ["weight", "hotspot_df", "AI", "svg_cluster"]:
            if getattr(self, attr) is not None:
                descr += f"{attr}, "
                flag = 1
        if flag:
            descr = descr[:-2]
        return descr

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, pos) -> STDataset:
        """
        Return a sub STDataset instance with empty attributes.
        """
        try:
            count_sub = self.count_df.loc[pos]
        except TypeError:
            count_sub = self.count_df.iloc[pos]
        coor_sub = self.coordinate_df.loc[count_sub.index, ]
        return STDataset(
            count_sub,
            coor_sub,
            check_duplicate_genes=False,
            sort_spots=False,
            make_sparse=pd.api.types.is_sparse(self.count_df.iloc[:, 1]),
        )

    def __del__(self) -> None:
        del self._count_df
        del self._coordinate_df
        del self._normalizer
        del self._hotspot_df
        del self._weight
        del self._weight_type
        del self._local_moran_i
        del self._local_moran_p
        del self._AI
        del self._Di
        del self._svg_cluster
        del self._spot_type
        del self._array_coordinate

    def to_dense(self) -> None:
        """Convert count_df with sparse values to dense."""
        self._count_df = self._count_df.sparse.to_dense()

    def to_sparse(self) -> None:
        """Convert count_df with dense values to sparse."""
        is_int = [i == "int" for i in self.count_df.dtypes]
        if all(is_int):
            dt = "int64"
        else:
            dt = "float64"
        self._count_df = self._count_df.astype(pd.SparseDtype(dt, 0))

    def acquire_weight(self, k: int = 6, **kwargs) -> None:
        """
        Acquire weight for analysis.

        Parameters
        ==========
        k : int, default 6
            Number of nearest neighbors for KNN network.

        **kwargs
            Additional keyword arguments passed to the libpysal.weights.KNN call.

        """
        self._weight = KNN(self._coordinate_df, k=k, **kwargs)
        self._weight_type = ("KNN", str(k))

    def acquire_hotspot(self, **kwargs) -> None:
        """
        Acquire hotspot matrix.

        Parameters
        ==========
        **kwargs
            Additional keyword arguments passed to local_moran call.

        """
        if self._weight is None:
            self.acquire_weight()
        hotspot, i_value, p_value = moran.local_moran(
            gene_expression_df=self.count_df,
            weights=self.weight,
            **kwargs,
        )
        self._hotspot_df = hotspot.reindex(
            index=self.spots,
            columns=self.genes,
        )
        self._local_moran_i = i_value.astype(pd.SparseDtype("float", 0))
        self._local_moran_p = p_value.astype(pd.SparseDtype("float", 0))

    def acquire_density(self, cores: int = density.cpu_count()) -> None:
        """
        Acquire local Di and global AI value.

        Parameters
        ==========
        cores : int
            Number of threads to run svgbit. Use all available cpus by default.

        """
        if self._hotspot_df is None:
            self.acquire_hotspot()
        results = density.hotspot_AI(
            hotspot_df=self._hotspot_df,
            weight_df=-np.log(self._local_moran_p),
            knn=self._weight,
            cores=cores,
        )
        self._AI = results[0].reindex(index=self.genes)
        self._Di = results[1].astype(pd.SparseDtype("float", 0)).reindex(
            index=self.spots, columns=self.genes)

    def find_clusters(
        self,
        n_svgs: int = 1000,
        n_svg_clusters: int = 8,
        threshold: float = 0.3,
    ) -> None:
        """
        Find SVG clusters.

        Parameters
        ==========
        n_svgs : int, default 1000
            Number of SVGs to find clusters.

        n_svg_clusters : int, default 8
            Number of SVG clusters to find.

        threshold : float, dafault 0.3
            min value to identify multiple svg clusters to spot.

        """
        results = cluster.cluster(
            self._hotspot_df,
            self._AI,
            n_svgs=n_svgs,
            n_svg_clusters=n_svg_clusters,
            threshold=threshold,
        )
        self._svg_cluster = results[0]
        self._spot_type = results[1].reindex(index=self.spots)

    @property
    def count_df(self) -> pd.DataFrame:
        """Expression matrix."""
        return self._count_df

    @property
    def coordinate_df(self) -> pd.DataFrame:
        """Coordinate information."""
        return self._coordinate_df

    @property
    def n_spots(self) -> int:
        """Number of total spots."""
        return self._count_df.shape[0]

    @property
    def spots(self) -> pd.Index:
        """An Index for spots' names."""
        return self._count_df.index

    @property
    def n_genes(self) -> int:
        """Number of total genes."""
        return self._count_df.shape[1]

    @property
    def genes(self) -> pd.Index:
        """An Index for genes' names."""
        return self._count_df.columns

    @property
    def weight(self) -> libpysal_W:
        """Weight used by svgbit. Use KNN if not specified."""
        return self._weight

    @weight.setter
    def weight(self, value: libpysal_W):
        self._weight = value
        self._weight_type = ("User specified weight", None)

    @property
    def weight_type(self) -> Tuple[Optional[str], Optional[str]]:
        """
        What kind of weight is used. The second element indicates parameter k
        used by KNN in default.
        """
        return self._weight_type

    @property
    def hotspot_df(self) -> pd.DataFrame:
        """Hotspot matrix."""
        return self._hotspot_df

    @property
    def AI(self) -> pd.Series:
        """A Series for AI value."""
        return self._AI

    @property
    def Di(self) -> pd.DataFrame:
        """A DataFrame for local Di value."""
        return self._Di

    @property
    def svg_cluster(self) -> pd.Series:
        """SVG cluster result."""
        return self._svg_cluster

    @property
    def spot_type(self) -> pd.DataFrame:
        """A pd.DataFrame for spot type."""
        return self._spot_type


if __name__ == "__main__":
    pass



from importlib.metadata import version, PackageNotFoundError

from .core.STDataset import STDataset
from .run import run

from .core.io import load_10X, load_anndata_h5, load_table
from .core.combinations import find_combinations
from . import filters, normalizers, plot

try:
    __version__ = version("svgbit")
except PackageNotFoundError:
    __version__ = "unknown version"

__all__ = [
    "STDataset",
    "run",
    "load_10X",
    "load_anndata_h5",
    "load_table",
    "filters",
    "normalizers",
    "plot",
    "find_combinations",
]



from __future__ import annotations

from numpy import log as nlog
from pandas import SparseDtype

from .STDataset import STDataset


def cpm_normalizer(dataset: STDataset) -> STDataset:
    """
    Perform CPM on dataset.

    Parameters
    ==========
    dataset : STDataset
        STDataset to be normalized.

    Returns
    =======
    dataset : STDataset
        A new STDataset instance with normalized expression matrix.

    """
    try:
        count_df = dataset.count_df.sparse.to_dense()
    except AttributeError:
        pass
    try:
        scale_df = (count_df.T * 10000 / count_df.T.sum())
    except AttributeError:
        pass
    return_dset = STDataset(
        scale_df.T,
        dataset.coordinate_df,
        check_duplicate_genes=False,
        sort_spots=False,
    )
    return_dset._normalizer = "cpm"
    return_dset._array_coordinate = dataset._array_coordinate
    return return_dset


def logcpm_normalizer(dataset: STDataset) -> STDataset:
    """
    Perform logcpm on dataset.

    Parameters
    ==========
    dataset : STDataset
        STDataset to be normalized.

    Returns
    =======
    dataset : STDataset
        A new STDataset instance with normalized expression matrix.

    """
    try:
        count_df = dataset.count_df.sparse.to_dense()
    except AttributeError:
        pass
    scale_df = nlog(count_df.T * 10000 / count_df.T.sum() + 1)
    return_dset = STDataset(
        scale_df.T,
        dataset.coordinate_df,
        check_duplicate_genes=False,
        sort_spots=False,
    )
    return_dset._normalizer = "logcpm"
    return_dset._array_coordinate = dataset._array_coordinate
    return return_dset



from __future__ import annotations

from multiprocessing import cpu_count

from .core.STDataset import STDataset


def run(
        dataset: STDataset,
        k: int = 6,
        n_svgs: int = 1000,
        n_svg_clusters: int = 8,
        cores: int = cpu_count(),
) -> STDataset:
    """
    Run all part of svgbit within one function.

    Parameters
    ==========
    dataset : STDataset
        A STDataset for running svgbit.

    k : int, default 6
        Number of nearest neighbors for KNN network.

    n_svgs : int, default 1000
        Number of SVGs to find clusters.

    n_svg_clusters : int, default 8
        Number of SVG clusters to find.

    cores : int
        Number of threads to run svgbit. Use all available cpus by default.

    Returns
    =======
    dataset : STDataset
        A STDataset with all evaluates done.

    """
    dataset.acquire_weight(k=k)
    dataset.acquire_hotspot(cores=cores)
    dataset.acquire_density(cores=cores)
    dataset.find_clusters(n_svgs=n_svgs, n_svg_clusters=n_svg_clusters)

    return dataset



from collections import Counter
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors
from sklearn.mixture import BayesianGaussianMixture

from .STDataset import STDataset


def _find_combinations(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    spot_type: pd.DataFrame,
    svg_cluster: pd.Series,
    center_spots: Union[int, list],
    selected_genes: Optional[list] = None,
    use_neighbor: bool = True,
) -> pd.DataFrame:
    """
    Find gene pairs in certain SVG cluster.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A hotspot DataFrame generated by svgbit.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    spot_type : pd.DataFrame
        A pd.DataFrame for assigned type_df.

    center_spots : int or list
        If a ``int`` is given, find gene pairs for this SVG cluster. If a
        ``list`` of spots is given, find gene pairs within those spots.
        ``selected_genes`` should be given if ``center_cluster`` is a ``list``.

    selected_genes : list or None, default None
        If a ``list`` of genes is given, find gene pairs within given genes.

    use_neighbor : bool, default True
        Whether to find gene pairs with neighbor SVG clusters.

    Returns
    =======
    gene_pairs_df : pd.DataFrame
        A pd.DataFrame for gene pairs in center_cluster with weights.
    """
    try:
        center_spots = list(center_spots)
    except TypeError:
        try:
            center_spots = int(center_spots)
        except TypeError:
            err = (f"center_spots must be a number or a list-like object, "
                   "not '{type(center_spots)}'")
            raise TypeError(err)

    if use_neighbor and (not isinstance(center_spots, list)):
        # find neighbor clusters
        certain_series = spot_type[spot_type["spot_type"] != "uncertain"]
        certain_series = certain_series["type_1"]
        nbrs = NearestNeighbors(n_neighbors=7).fit(coordinate_df)
        neighbor_clusters_ = {}
        for present_cluster in set(spot_type["type_1"]):
            used_spots = []
            neighbor_clusters_[present_cluster] = []
            present_spots = certain_series[certain_series ==
                                           present_cluster].index
            distances, indices = nbrs.kneighbors(
                coordinate_df.reindex(index=present_spots))
            for record in indices:
                for neighbor_spot in record[1:]:
                    if neighbor_spot in used_spots:
                        continue
                    used_spots.append(neighbor_spot)
                    neighbor_cluster = spot_type.iloc[neighbor_spot, 1]
                    if neighbor_cluster == present_cluster:
                        continue
                    neighbor_clusters_[present_cluster].append(
                        neighbor_cluster)

        neighbor_clusters = {}
        for i in neighbor_clusters_:
            counter = Counter(neighbor_clusters_[i])
            coverage_ = len(certain_series[certain_series == i]) * 0.03
            neighbor_clusters[i] = [
                j[0] for j in counter.most_common()[:3] if j[1] >= coverage_
            ]
            neighbor_clusters[i].append(i)

        selected_spots = [
            j for i in neighbor_clusters[center_spots]
            for j in spot_type[spot_type["type_1"] == i].index
        ]
        selected_genes = [
            j for i in neighbor_clusters[center_spots]
            for j in svg_cluster[svg_cluster == i].index
        ]
    else:
        if isinstance(center_spots, list):
            selected_spots = center_spots
            if len(center_spots) <= 10:
                center_spots = ", ".join(center_spots)
            else:
                center_spots = ", ".join(center_spots[:3])
                center_spots = center_spots + "..."
        else:
            selected_spots = spot_type[spot_type["type_1"] ==
                                       center_spots].index
            if selected_genes is None:
                selected_genes = svg_cluster[svg_cluster == center_spots].index

    count_sub = hotspot_df.reindex(
        index=selected_spots,
        columns=selected_genes,
    )
    con_matrix = count_sub.T @ count_sub
    n_hotspots = np.diag(con_matrix.to_numpy())
    con_matrix = con_matrix - np.diag(n_hotspots)
    men_matrix = (1 - count_sub.T) @ count_sub

    con_ratio = con_matrix / len(selected_spots)
    men_ratio = men_matrix / len(selected_spots)

    gene_pairs_df = pd.DataFrame(columns=[
        "SVG_cluster", "gene_1", "gene_2", "colocalization_score",
        "exclusive_score"
    ])
    line = 0
    for gene_1 in con_ratio.columns:
        for gene_2 in con_ratio.index:
            if gene_1 == gene_2:
                continue
            write_dict = {
                "SVG_cluster": center_spots,
                "gene_1": gene_1,
                "gene_2": gene_2,
                "colocalization_score": con_ratio[gene_1][gene_2],
                "exclusive_score": men_ratio[gene_1][gene_2],
            }
            write_series = pd.Series(write_dict, name=line).to_frame().T
            line += 1
            gene_pairs_df = pd.concat([gene_pairs_df, write_series])

    gene_pairs_df.fillna(0, inplace=True)

    gmm_con = BayesianGaussianMixture(
        n_components=3,
        max_iter=500,
        n_init=5,
    )
    result_con = gmm_con.fit_predict(
        gene_pairs_df["colocalization_score"].to_numpy().reshape(-1, 1))
    rank_str = ["low", "middle", "high"]
    mean_argsort = gmm_con.means_.argsort(axis=0).flatten()
    rank_dict = {i: j for i, j in zip(mean_argsort, rank_str)}
    result_con = [rank_dict[i] for i in result_con]
    gene_pairs_df["colocalization_degree"] = result_con

    gmm_men = BayesianGaussianMixture(
        n_components=3,
        max_iter=500,
        n_init=5,
    )
    result_men = gmm_men.fit_predict(
        gene_pairs_df["exclusive_score"].to_numpy().reshape(-1, 1))
    mean_argsort = gmm_men.means_.argsort(axis=0).flatten()
    rank_dict = {i: j for i, j in zip(mean_argsort, rank_str)}
    result_men = [rank_dict[i] for i in result_men]
    gene_pairs_df["exclusive_degree"] = result_men

    return gene_pairs_df


def find_combinations(
    dataset: STDataset,
    center_spots: Union[int, list],
    selected_genes: Optional[list] = None,
    use_neighbor: bool = True,
) -> pd.DataFrame:
    """
    Find gene pairs in certain SVG cluster.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with all steps finished.

    center_spots : int or list
        If a ``int`` is given, find gene pairs for this SVG cluster. If a
        ``list`` of spots is given, find gene pairs within those spots.
        ``selected_genes`` should be given if ``center_cluster`` is a ``list``.

    selected_genes : list or None, default None
        If a ``list`` of genes is given, find gene pairs within given genes.

    use_neighbor : bool, default True
        Whether to find gene pairs with neighbor SVG clusters.

    Returns
    =======
    gene_pairs_df : pd.DataFrame
        A pd.DataFrame for gene pairs in center_cluster with weights.
    """
    return _find_combinations(
        dataset.hotspot_df,
        dataset.coordinate_df,
        dataset.spot_type,
        dataset.svg_cluster,
        center_spots,
        selected_genes,
        use_neighbor,
    )



from .core.plot import svg_heatmap
from .core.plot import hotspot_distribution_map
from .core.plot import spot_type_map
from .core.plot import hotspot_expression, gene_expression
from .core.plot import hotspot_colocalization_map



from __future__ import annotations

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from libpysal.weights import W as libpysal_W
from scipy.stats import norm


def pysal_to_pandas(W: libpysal_W) -> pd.DataFrame:
    return pd.DataFrame(*W.full()).astype(int).T


def plot_gmm(gmm, hist_data, save_path) -> None:
    """Plot a GMM given by sklearn."""
    fig, ax = plt.subplots(figsize=(10, 10))
    ax.hist(hist_data, bins=100, density=True, color="#f0a1a8")
    x = np.linspace(min(hist_data), max(hist_data), 5000)
    ax.plot(
        x,
        np.exp(gmm.score_samples(x.reshape(-1, 1))),
        lw=4,
        label="GMM",
    )
    for i in range(gmm.get_params()["n_components"]):
        ax.plot(
            x,
            norm.pdf(
                x,
                gmm.means_[i, 0],
                gmm.covariances_[i, 0]**(1 / 2),
            ) * gmm.weights_[i],
            lw=4,
            ls="--",
            label=f"Gaussian {i}, weight {gmm.weights_[i]:.3f}",
        )
    ax.legend()
    fig.savefig(save_path)


def get_cmap(length):
    return plt.colormaps["hsv"].resampled(length)



from __future__ import annotations

import gzip
from pathlib import Path

import anndata
import h5py
import numpy as np
import pandas as pd
import scipy.io
from .STDataset import STDataset


def load_10X(read_path, make_sparse=True) -> STDataset:
    """
    Load 10X Genomics Space Ranger outputs and generate STDataset.

    Parameters
    ==========
    read_path : str or pathlib.Path
        A location points to 10X outs dir. Assume directories
        ``filtered_feature_bc_matrix`` and ``spatial`` are in this
        path.

    make_sparse : bool, default True
        Whether to use sparse DataFrame in order to save memory.

    Returns
    =======
    dataset : STDataset
        A STDataset instance generated from read_path.
    """
    read_path = Path(read_path)
    mat_dir = Path.joinpath(read_path, "filtered_feature_bc_matrix")
    mtx_path = Path.joinpath(mat_dir, "matrix.mtx.gz")
    features_path = Path.joinpath(mat_dir, "features.tsv.gz")
    barcodes_path = Path.joinpath(mat_dir, "barcodes.tsv.gz")
    position_path = Path.joinpath(read_path, "spatial",
                                  "tissue_positions_list.csv")

    gene_name = []
    with gzip.open(features_path, "rt") as f:
        for line in f:
            line = line.strip()
            gene_name.append(line.split("\t")[1])

    spot_name = []
    with gzip.open(barcodes_path, "rt") as f:
        for line in f:
            line = line.strip()
            spot_name.append(line)

    count_df = pd.DataFrame.sparse.from_spmatrix(scipy.io.mmread(mtx_path), ).T
    count_df.index = spot_name
    count_df.columns = gene_name

    try:
        coor_df = pd.read_csv(position_path, index_col=0, header=None)
        spaceranger_version = "v1"
    except FileNotFoundError:
        position_path = Path.joinpath(read_path, "spatial",
                                      "tissue_positions.csv")
        coor_df = pd.read_csv(position_path, index_col=0, header=0)
        spaceranger_version = "v2"
    if spaceranger_version == "v1":
        array_coor = coor_df[[3, 2]]
        coor_df = coor_df[[5, 4]]
    elif spaceranger_version == "v2":
        array_coor = coor_df[["array_row", "array_col"]]
        coor_df = coor_df[["pxl_row_in_fullres", "pxl_col_in_fullres"]]
    coor_df.index.name = "barcode"
    coor_df.columns = ["X", "Y"]
    array_coor = array_coor.reindex(index=count_df.index)
    array_coor.columns = ["X", "Y"]
    coor_df = coor_df.reindex(index=count_df.index)

    if not make_sparse:
        count_df = count_df.sparse.to_dense()
    dataset = STDataset(count_df, coor_df, make_sparse=make_sparse)
    dataset._array_coordinate = array_coor
    return dataset


def load_anndata_h5(read_path, **kwargs) -> STDataset:
    """
    Load anndata saved h5ad file and generate STDataset.

    .. note::
        ``load_anndata_h5`` will try to use anndata.X as expression matrix.

    Parameters
    ==========
    read_path : str or pathlib.Path
        File name to read from.

    **kwargs
        Additional keyword arguments passed to anndata.read_h5ad

    Returns
    =======
    dataset : STDataset
        A STDataset instance generated from read_path.
    """
    adata = anndata.read_h5ad(read_path, **kwargs)
    count_df = pd.DataFrame.sparse.from_spmatrix(
        adata.X,
        index=adata.obs.index,
        columns=adata.var.index,
    )
    coor_df = pd.DataFrame(adata.obsm["spatial"])
    coor_df.index = count_df.index
    coor_df.columns = ["X", "Y"]

    if isinstance(count_df.iloc[0, 0], np.float32):
        count_df = count_df.astype(np.float64)

    dataset = STDataset(count_df, coor_df)
    return dataset


def load_table(read_path, **kwargs) -> STDataset:
    """
    Load text file and generate STDataset.

    Support tables in following format:

    ======== === === ======== ============
     geneID   X   Y   counts   spot_name
                               (optional)
    ======== === === ======== ============
     gene_1   1   1      1      spot_1
     gene_2   1   1      2      spot_1
     gene_1   1   2      1      spot_2
     gene_3   1   2      3      spot_2
      ...    ... ...    ...       ...
    ======== === === ======== ============

    Parameters
    ==========
    read_path : str or pathlib.Path
        File name to read from.

    **kwargs
        Additional keyword arguments passed to pd.read_csv

    Returns
    =======
    dataset : STDataset
        A STDataset instance generated from read_dir.
    """
    read_df = pd.read_csv(read_path, **kwargs)
    count_dict = {}
    coor_dict = {}
    for it in read_df.iterrows():
        try:
            spot_name = it[1].iloc[3]
        except IndexError:
            spot_name = f"{it[1].iloc[0]}x{it[1].iloc[1]}"
        if spot_name not in coor_dict.keys():
            coor_dict[spot_name] = {"X": it[1].iloc[0], "Y": it[1].iloc[1]}
        if spot_name not in count_dict.keys():
            count_dict[spot_name] = {}
        if it[0] not in count_dict[spot_name].keys():
            count_dict[spot_name][it[0]] = it[1][2]
        else:
            count_dict[spot_name][it[0]] += it[1][2]
    count_df = pd.DataFrame.from_dict(count_dict).fillna(0).T
    coor_df = pd.DataFrame.from_dict(coor_dict).T
    coor_df = coor_df.reindex(index=count_df.index)

    return STDataset(count_df, coor_df)


def load_gef(read_path, slot="bin20") -> STDataset:
    """
    Load gef file and generate STDataset.

    Parameters
    ==========
    read_path : str or pathlib.Path
        File name to read from.

    Returns
    =======
    dataset : STDataset
        A STDataset instance generated from read_dir.
    """
    count_dict = {}
    coor_dict = {}
    f = h5py.File(read_path)
    for record in f["geneExp"][slot]["gene"]:
        gene_symbol = record[0].decode("utf8")
        begin_line = record[1]
        end_line = record[1] + record[2] - 1
        exp = f["geneExp"][slot]["expression"][begin_line:end_line]
        for line in exp:
            line = list(map(int, line))
            spot_name = f"{line[0]}x{line[1]}"
            if spot_name not in coor_dict.keys():
                coor_dict[spot_name] = {"X": line[0], "Y": line[1]}
            if spot_name not in count_dict.keys():
                count_dict[spot_name] = {}
            if gene_symbol not in count_dict[spot_name].keys():
                count_dict[spot_name][gene_symbol] = line[2]
            else:
                count_dict[spot_name][gene_symbol] += line[2]
    count_df = pd.DataFrame.from_dict(count_dict).fillna(0).T
    coor_df = pd.DataFrame.from_dict(coor_dict).T
    coor_df = coor_df.reindex(index=count_df.index)
    f.close()

    return STDataset(count_df, coor_df)



from __future__ import annotations

import pandas as pd
from scipy.cluster import hierarchy as sch


def cluster(
    hotspot_df: pd.DataFrame,
    AI_series: pd.Series,
    n_svgs: int = 1000,
    n_svg_clusters: int = 8,
    threshold: float = 0.3,
) -> pd.Series:
    """
    Cluster SVGs using hotspot matrix.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A hotspot DataFrame generated by svgbit.

    AI_series : pd.Series
        A Series for AI value.

    n_svgs : int, default 1000
        Number of SVGs to find clusters.

    n_svg_clusters : int, default 8
        Number of SVG clusters to find.

    threshold : float, dafault 0.3
        min value to identify multiple svg clusters to spot.

    Returns
    =======
    gene_result : pd.Series
        A Series of clustering result.

    """
    selected_genes = AI_series.sort_values(ascending=False)[:n_svgs].index
    hotspot_set = hotspot_df[selected_genes]
    gene_distmat = sch.distance.pdist(hotspot_set.T, metric="jaccard")
    Z_gene = sch.linkage(gene_distmat, method="ward")
    gene_result = pd.Series(
        sch.fcluster(Z_gene, t=n_svg_clusters, criterion="maxclust"),
        index=selected_genes,
    ).sort_values()

    mean_df = pd.DataFrame(index=hotspot_df.index)
    for i in range(1, n_svg_clusters + 1):
        genes = gene_result[gene_result == i].index
        mean_df[i] = hotspot_df[genes].T.mean()

    columns = ["spot_type"]
    [columns.append(f"type_{i}") for i in range(1, 1 + n_svg_clusters)]
    type_df = pd.DataFrame(columns=columns)
    for spot in mean_df.index:
        temp_df = pd.DataFrame(index=[spot], columns=columns)
        sort_series = mean_df.loc[spot, :].sort_values(ascending=False)
        n_multi = len(sort_series[sort_series > threshold])
        spot_type = "singlet"
        if sort_series.iloc[0] < 0.2:
            spot_type = "uncertain"
        if n_multi > 1:
            spot_type = f"{n_multi}_multi_types"
        temp_df.loc[spot, "spot_type"] = spot_type
        temp_df.loc[spot, columns[1:]] = sort_series.index
        type_df = pd.concat([type_df, temp_df])

    return gene_result, type_df



from __future__ import annotations

from itertools import combinations
from math import ceil
from pathlib import Path
from typing import Optional, Union, List

import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
from PIL import Image
from scipy.special import comb

from .STDataset import STDataset
from .utils import get_cmap


def _svg_heatmap(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    cluster_result: pd.Series,
    spot_type: pd.DataFrame,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw SVG distribution heatmap.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    cluster_result : pd.Series
        A pd.Series for cluster result.

    spot_type : pd.DataFrame
        A pd.DataFrame for assigned type_df.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    spot_type = spot_type.sort_values(by="type_1")

    left, bottom = 0.1, 0.1
    width, height = 0.66, 0.66
    spacing, cluster_width = 0.03, 0.22
    rect_heatmap = [left + spacing, bottom + spacing * 2, width, height]

    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax_heatmap = fig.add_axes(rect_heatmap)

    for i, j in enumerate(set(cluster_result.values)):
        rect_cluster = [
            left + width + spacing * (2 + i // 3) + cluster_width * (i // 3),
            bottom + spacing * (3 - i % 3) + cluster_width * (2 - i % 3),
            cluster_width,
            cluster_width,
        ]
        ax_cluster = fig.add_axes(rect_cluster)
        ax_cluster.set_title(f"Cluster {j} hotspot distribution")
        ax_cluster.set_xticks([])
        ax_cluster.set_yticks([])
        ax_cluster.imshow(he_image) if he_image is not None else None
        ax_cluster.axis("off")
        sc = ax_cluster.scatter(
            coordinate_df.iloc[:, 0],
            coordinate_df.iloc[:, 1],
            c=hotspot_df[cluster_result[cluster_result == j].index].T.mean(),
            cmap="autumn_r",
            vmin=0,
            vmax=1,
            s=s,
            alpha=0.7,
        )

    ii = ceil(len(set(cluster_result.values)) / 3)
    rect_cb = [
        left + width + spacing * (ii + 2) + cluster_width * ii,
        bottom + spacing * 2,
        0.02,
        height,
    ]
    ax_cb = fig.add_axes(rect_cb)
    fig.colorbar(sc, cax=ax_cb)

    ax_heatmap.imshow(
        hotspot_df.reindex(
            index=spot_type.index,
            columns=cluster_result.index,
        ),
        cmap="Reds",
        aspect="auto",
    )

    flag = 0
    for i in range(1, len(set(cluster_result.values))):
        flag += len(cluster_result[cluster_result == i])
        ax_heatmap.plot([flag, flag], [0, hotspot_df.shape[0] - 5])

    ax_heatmap.set_xticks([])
    ax_heatmap.set_yticks([])
    ax_heatmap.set_xlabel("Genes")
    ax_heatmap.set_ylabel("Spots")

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def _hotspot_distribution_map(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    cluster_result: pd.Series,
    cluster: Union[str, int],
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw hotspot distribution map for one SVG cluster.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    cluster_result : pd.Series
        A pd.Series for cluster result.

    cluster : str or int
        Specify drawing cluster.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    cluster = int(cluster)

    ax.set_title(f"Cluster {cluster} hotspot distribution")
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis("off")
    sc = ax.scatter(
        coordinate_df.iloc[:, 0],
        coordinate_df.iloc[:, 1],
        c=hotspot_df[cluster_result[cluster_result == cluster].index].T.mean(),
        cmap="autumn_r",
        vmin=0,
        vmax=1,
        s=s,
        alpha=0.7,
    )
    fig.colorbar(sc)

    ax.imshow(he_image) if he_image is not None else None

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def _hotspot_expression(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    gene: str,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw hotspot expression for one gene.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    gene : str
        Which gene to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    ax.set_title(gene)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis("off")
    sc = ax.scatter(
        coordinate_df.iloc[:, 0],
        coordinate_df.iloc[:, 1],
        c=hotspot_df[gene],
        cmap="autumn_r",
        vmin=0,
        vmax=1,
        s=s,
        alpha=0.7,
    )
    fig.colorbar(sc)

    ax.imshow(he_image) if he_image is not None else None

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def _hotspot_colocalization_map(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    genes: list,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
    colors: Optional[List] = None,
) -> mpl.figure.Figure:
    """
    Draw hotspot expression for multiple genes.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    genes : str
        Which genes to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    colors : list, default None
        A list for spot colors. If None, auto generate colors.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    if colors is None:
        if len(genes) <= 3:
            colors = [
                "tab:red", "tab:green", "tab:blue", "tab:orange", "tab:pink",
                "tab:cyan", "k"
            ]
        else:
            i = len(genes)
            n_colors = 0
            while i > 0:
                n_colors += comb(len(genes), i)
                i -= 1
            colors = get_cmap(n_colors)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis("off")

    for i, gene in enumerate(genes):
        draw_counts = hotspot_df[gene].loc[hotspot_df[gene] == 1]
        ax.scatter(
            coordinate_df.iloc[:, 0].reindex(index=draw_counts.index),
            coordinate_df.iloc[:, 1].reindex(index=draw_counts.index),
            s=s,
            color=colors[i],
            alpha=0.7,
            label=f"{gene}",
        )

    n_combinations = 2
    while n_combinations <= len(genes):
        for c in combinations(genes, n_combinations):
            i += 1
            draw_counts = pd.Series(index=coordinate_df.index, dtype=int)
            draw_counts = draw_counts.fillna(0)
            for gene in c:
                draw_counts += hotspot_df[gene]
            draw_counts = draw_counts[draw_counts == n_combinations]
            ax.scatter(
                coordinate_df.iloc[:, 0].reindex(index=draw_counts.index),
                coordinate_df.iloc[:, 1].reindex(index=draw_counts.index),
                c=colors[i],
                s=s,
                label=" & ".join(c),
            )
        n_combinations += 1

    ax.legend(markerscale=2)

    ax.imshow(he_image) if he_image is not None else None

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def _gene_expression(
    expression_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    gene: str,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw expression for one gene.

    Parameters
    ==========
    count_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    gene : str
        Which gene to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    ax.set_title(gene)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis("off")
    sc = ax.scatter(
        coordinate_df.iloc[:, 0],
        coordinate_df.iloc[:, 1],
        c=expression_df[gene],
        cmap="autumn_r",
        s=s,
        alpha=0.7,
    )
    fig.colorbar(sc)

    ax.imshow(he_image) if he_image is not None else None

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def _spot_type_map(
    hotspot_df: pd.DataFrame,
    coordinate_df: pd.DataFrame,
    type_df: pd.DataFrame,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    draw_uncertain: bool = True,
    s: float = 16,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw SVG type map.

    Parameters
    ==========
    hotspot_df : pd.DataFrame
        A pd.DataFrame for hotspots.

    coordinate_df : pd.DataFrame
        A pd.DataFrame for coordinate files.

    type_df : pd.DataFrame
        A pd.DataFrame for spot type.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    draw_uncertain : bool, default True
        Whether to draw uncertain spots.

    s : float, default 16
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """

    if draw_uncertain:
        certain_spots = type_df.index
    else:
        certain_spots = type_df[type_df["spot_type"] != "uncertain"].index
    coordinate_df = coordinate_df.reindex(index=certain_spots)
    type_df = type_df.reindex(index=certain_spots)

    ncs = len(set(type_df["type_1"]))
    if ncs <= 10:
        cmap = "tab10"
        ncol = 1
    elif ncs <= 20:
        cmap = "tab20"
        ncol = 2
    else:
        cmap = get_cmap(ncs)
        ncol = 4

    if mpl.rcParams["legend.title_fontsize"] is None:
        legend_fontsize = 16
    else:
        legend_fontsize = mpl.rcParams["legend.title_fontsize"]

    he_close = False
    figsize = (10, 10)
    if he_image is not None:
        if not isinstance(he_image, Image.Image):
            he_close = True
            he_image = Image.open(he_image)
            figsize = (10, he_image.height / he_image.width * 10)

    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    sc = ax.scatter(
        coordinate_df["X"],
        coordinate_df["Y"],
        s=s,
        c=type_df["type_1"],
        cmap=cmap,
    )
    leg = ax.legend(
        *sc.legend_elements(),
        ncol=ncol,
        title="Cluster",
        title_fontsize=legend_fontsize,
        fontsize=legend_fontsize,
    )
    ax.add_artist(leg)

    ax.axis("off")
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title("Spot type map", fontsize=legend_fontsize)

    ax.imshow(he_image) if he_image is not None else None

    if save_path is not None:
        fig.savefig(save_path, bbox_inches="tight")

    he_image.close() if he_close else None

    return fig


def svg_heatmap(
    dataset: STDataset,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw SVG distribution heatmap.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot and SVG cluster estimation finished.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _svg_heatmap(
        hotspot_df=dataset.hotspot_df,
        coordinate_df=coor_df,
        cluster_result=dataset.svg_cluster,
        spot_type=dataset.spot_type,
        save_path=save_path,
        he_image=he_image,
        s=s,
        dpi=dpi,
    )


def hotspot_distribution_map(
    dataset: STDataset,
    cluster: Union[str, int],
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw hotspot distribution map for one SVG cluster.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot and SVG cluster estimation finished.

    cluster : str or int
        Specify drawing cluster.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _hotspot_distribution_map(
        hotspot_df=dataset.hotspot_df,
        coordinate_df=coor_df,
        cluster_result=dataset.svg_cluster,
        cluster=cluster,
        save_path=save_path,
        he_image=he_image,
        s=s,
        dpi=dpi,
    )


def hotspot_expression(
    dataset: STDataset,
    gene: str,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw hotspot expression for one gene.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot estimation finished.

    gene : str
        Which gene to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _hotspot_expression(
        hotspot_df=dataset.hotspot_df,
        coordinate_df=coor_df,
        gene=gene,
        save_path=save_path,
        he_image=he_image,
        s=s,
        dpi=dpi,
    )


def hotspot_colocalization_map(
    dataset: STDataset,
    genes: list,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
    colors: Optional[list] = None,
) -> mpl.figure.Figure:
    """
    Draw hotspot expression for multiple genes.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot estimation finished.

    genes : str
        Which genes to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    colors : list, default None
        A list for spot colors. If None, auto generate colors.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _hotspot_colocalization_map(
        hotspot_df=dataset.hotspot_df,
        coordinate_df=coor_df,
        genes=genes,
        save_path=save_path,
        he_image=he_image,
        s=s,
        dpi=dpi,
        colors=colors,
    )


def gene_expression(
    dataset: STDataset,
    gene: str,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    s: float = 4,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw expression for one gene.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot and SVG cluster estimation finished.

    gene : str
        Which gene to draw.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    s : float, default 4
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _gene_expression(
        expression_df=dataset.count_df,
        coordinate_df=coor_df,
        gene=gene,
        save_path=save_path,
        he_image=he_image,
        s=s,
        dpi=dpi,
    )


def spot_type_map(
    dataset: STDataset,
    save_path: Optional[Union[str, Path]] = None,
    he_image: Optional[Union[str, Path, Image.Image]] = None,
    draw_uncertain: bool = True,
    s: float = 16,
    dpi: float = 300,
) -> mpl.figure.Figure:
    """
    Draw SVG type map.

    Parameters
    ==========
    dataset : STDataset
        A STDataset with hotspot and SVG cluster estimation finished.

    save_path : str or pathlib.Path, default None
        Heatmap save path. If None, plot will not save.

    he_image : str, pathlib.Path or PIL.Image.Image, default None
        H&E image of tissue. If None is given (default), distribution map
        will not show tissue picture.

    draw_uncertain : bool, default True
        Whether to draw uncertain spots.

    s : float, default 16
        Spot size.

    dpi : float, default 300
        DPI for saved figure.

    Returns
    =======
    fig : matplotlib.figure.Figure
        A matplotlib.figure.Figure plot.
    """
    coor_df = dataset.coordinate_df
    if he_image is None:
        if dataset._array_coordinate is not None:
            coor_df = dataset._array_coordinate
    return _spot_type_map(
        hotspot_df=dataset.hotspot_df,
        coordinate_df=coor_df,
        type_df=dataset.spot_type,
        save_path=save_path,
        he_image=he_image,
        draw_uncertain=draw_uncertain,
        s=s,
        dpi=dpi,
    )
